﻿/// ==============================================================================================
/// Namespace   : ScipBe.Common.Geocoding
/// Class(es)   : GoogleGeoCoder
/// Version     : 1.1
/// Author      : Stefan Cruysberghs
/// Website     : http://www.scip.be
/// Date        : June 2008 - December 2008
/// Status      : Open source - Mozilla Public License 1.1
/// ==============================================================================================

using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Net;
using System.Xml.Linq;

namespace ScipBe.Common.Geocoding
{
  /// <summary>
  /// XDocument Helper class
  /// </summary>
  internal static class XDocumentHelper
  {
    /// <summary>
    /// Parse Xml element and get value from one of the given elements and attribute
    /// </summary>
    /// <param name="parentElement">Parent element</param>
    /// <param name="xmlElementPaths">Array of element paths</param>
    /// <param name="xmlAttributeName">Name of attribute</param>
    /// <returns>String value of first element in list which returns a not empty value</returns>
    public static string GetValue(XElement parentElement, string[] xmlElementPaths, string xmlAttributeName)
    {
      foreach(string path in xmlElementPaths)
      {
        string value = GetValue(parentElement, path, xmlAttributeName);
        if (value.Trim() != "")
          return value;
      }
      return "";
    }
    
    /// <summary>
    /// Parse Xml element and get value from given element and attribute
    /// </summary>
    /// <param name="parentElement">Parent element</param>
    /// <param name="xmlElementPath">Path of element</param>
    /// <param name="xmlAttributeName">Name of attribute</param>
    /// <returns>String value of given element or attribute</returns>
    public static string GetValue(XElement parentElement, string xmlElementPath, string xmlAttributeName)
    {
      string[] elementNames = xmlElementPath.Split('/');
      XElement element = parentElement;

      foreach (string name in elementNames.Where(n => n != ""))
      {
        element = element.Element(name);
        if (element == null)
        {
          return "";
        }
      }

      if (string.IsNullOrEmpty(xmlAttributeName))
      {
        return element.Value;
      }

      XAttribute attr = element.Attribute(xmlAttributeName);
      if (attr == null)
      {
        return "";
      }
      return attr.Value;
    }    
  }

  /// <summary>
  /// Google Geocoder class
  /// </summary>
  public class GoogleGeoCoder
  {
    /// <summary>
    /// Google Maps API key
    /// </summary>
    /// <remarks>
    /// http://code.google.com/apis/maps/signup.html 
    /// </remarks>
    private readonly string googleMapsKey;

    /// <summary>
    /// Constructor of GoogleGeocoder.
    /// </summary>
    /// <param name="key">Pass your Google Maps API key to get access to the Google services</param>
    public GoogleGeoCoder(string key)
    {
      googleMapsKey = key;
    }

    /// <summary>
    /// GeoListCompleted delegate
    /// </summary>
    /// <param name="sender">Sender</param>
    /// <param name="geoList">Collection of IGeoData objects</param>
    public delegate void GeoListCompletedEventHandler(object sender, IEnumerable<IGeoData> geoList);
    /// <summary>
    /// GeoListCompleted event. This will be triggered after calling GetGeoListAsync()
    /// </summary>
    public event GeoListCompletedEventHandler GeoListCompleted; 

    /// <summary>
    /// Google Maps geocoding service via HTTP. Pass address to Google Maps server and 
    /// return a simple IGeoData object which contains the coordinates like longitude and latitude
    /// </summary>
    /// <remarks>
    /// http://code.google.com/apis/maps/documentation/services.html#Geocoding_Direct
    /// </remarks>
    /// <param name="address">String with as many address parameters as you want (street, housenumber, postcode, city, country, ...)</param>
    /// <returns>GeoData object which contains Status, Accuracy, Longitude and Latitude</returns>
    /// <example><code>
    /// GoogleGeoCoder geoCoder = new GoogleGeoCoder("My key");
    /// IGeoData geoData = geoCoder.GetGeoData("Street City Country");
    /// if (geoData.Status == GeoStatusCode.Success)
    /// {
    ///   Console.WriteLine(geoData.Accuracy);
    ///   Console.WriteLine("{0}:{1}", geoData.Coordinate.Longitude, geoData.Coordinate.Latitude);
    /// }
    /// </code></example>
    public IGeoData GetGeoData(string address)
    {
      WebClient webClient = new WebClient();

      // Fill in arguments of the URI
      Uri baseAddressUri = new Uri("http://maps.google.com/");
      UriTemplate uriTemplate = new UriTemplate("maps/geo?q={address}&output={output}&key={key}");
      Uri formattedUri = uriTemplate.BindByPosition(baseAddressUri, address, "csv", googleMapsKey);

      // Call REST service, download CSV and save it as an array
      string[] geoDataParams = webClient.DownloadString(formattedUri).Split(',');

      // Google Maps always returns floats with a dot as decimal separator
      NumberFormatInfo provider = new NumberFormatInfo {NumberDecimalSeparator = "."};

      /* A reply returned in the CSV format consists of four numbers, separated by commas
       * 1 : HTTP status code
       * 2 : Accuracy 
       * 3 : Latitude 
       * 4 : Longitude
      */

      // Create GeoData class and initialize all properties
      return new GeoData
      {
        SearchAddress = address,
        Status = (GeoStatusCode)Convert.ToInt16(geoDataParams[0]),
        Accuracy = (GeoAddressAccuracy)Convert.ToInt16(geoDataParams[1]),
        Coordinate = new Coordinate()
        {
          Latitude = Convert.ToSingle(geoDataParams[2], provider),
          Longitude = Convert.ToSingle(geoDataParams[3], provider)
        }
      };
    }

    /// <summary>
    /// Call Google Maps geocoding service via HTTP. Pass address to Google Maps server and 
    /// return a collection of IGeoData objects which contain the coordinates and a lot of other geo data.
    /// </summary>
    /// <remarks>
    /// http://code.google.com/apis/maps/documentation/services.html#Geocoding_Direct
    /// http://mapki.com/wiki/Google_Map_Parameters
    /// </remarks>
    /// <param name="address">String with as many address parameters as you want (street, housenumber, postcode, city, country, ...)</param>
    /// <returns>Collection of IGeoData objects which contains Status, Accuracy, Longitude and Latitude, CountryNameCode,
    /// AdministrativeAreaName, PremiseName, PostalCodeNumber, ...</returns>
    /// <example><code>
    /// GoogleGeoCoder geoCoder = new GoogleGeoCoder("A Google Maps API key");
    /// IEnumerable&gt;IGeoData&lt; geoList = geoCoder.GetGeoList("Street City Country");
    /// if ((geoList.Count() > 0) &amp;&amp; (geoList.First().Status == GeoStatusCode.Success))
    /// {
    ///   foreach (var geoData in geoList)
    ///   {
    ///     Console.WriteLine("Address = {0}", geoData.Address);
    ///     Console.WriteLine("Thoroughfare = {0}", geoData.ThoroughfareName);
    ///     Console.WriteLine("Locality = {0}", geoData.DependentLocalityName);
    ///     Console.WriteLine("Administrative area = {0}", geoData.AdministrativeAreaName);
    ///     Console.WriteLine("Country = {0}", geoData.CountryNameCode);
    ///     Console.WriteLine("Coordinates = {0}:{1}", geoData.Coordinate.Longitude, geoData.Coordinate.Latitude);
    ///   } 
    /// }
    /// </code></example>
    public IEnumerable<IGeoData> GetGeoList(string address)
    {
      // Fill in arguments of the URI
      Uri baseAddressUri = new Uri("http://maps.google.com/");
      UriTemplate uriTemplate = new UriTemplate("maps/geo?q={address}&output={output}&key={key}");
      Uri formattedUri = uriTemplate.BindByPosition(baseAddressUri, address, "kml", googleMapsKey);

      // Call REST service and download XML
      WebClient webClient = new WebClient();
      string geoDataXml = webClient.DownloadString(formattedUri);
      return ProcessGeoDataXml(geoDataXml, address);
    }

    /// <summary>
    /// Process returned Xml data
    /// </summary>
    /// <param name="geoDataXml">Xml data which is returned by Google Maps server</param>
    /// <param name="address">Orignal search address</param>
    /// <returns>Collection of IGeoData objects which contains Status, Accuracy, Longitude and Latitude, CountryNameCode,
    /// AdministrativeAreaName, PremiseName, PostalCodeNumber, ...</returns>
    private static IEnumerable<IGeoData> ProcessGeoDataXml(string geoDataXml, string address)
    {
      // Remove the XML namespaces. This is needed to use LINQ to XML
      geoDataXml = geoDataXml.Replace(@"<kml xmlns=""http://earth.google.com/kml/2.0"">", "");
      geoDataXml = geoDataXml.Replace("</kml>", "");
      geoDataXml = geoDataXml.Replace(@"xmlns=""urn:oasis:names:tc:ciq:xsdschema:xAL:2.0""", "");
      XDocument xDoc = XDocument.Parse(geoDataXml);
      
      // Check status of returned data
      GeoStatusCode status = (GeoStatusCode)Convert.ToInt16(XDocumentHelper.GetValue(xDoc.Element("Response"), "Status/code", ""));

      // If not successful, then return one simple GeoData object
      if (status != GeoStatusCode.Success)
      {
        return new List<IGeoData>() { new GeoData() {SearchAddress = address, Status = status} };
      }

      // Google Maps always returns floats with a dot as decimal separator
      NumberFormatInfo provider = new NumberFormatInfo {NumberDecimalSeparator = "."};

      // LINQ to XML query to parse XML file and return list of placemarks
      // Use the XDocumentHelper class to try to get element and attribute values
      // If not element or attribute is not found, the GetValue() method will return an empty string
      // The XML elements ThoroughfareName, PremiseName and PostalCodeNumber can be stored in 4 different places
      var geoList =
        from item in xDoc.Descendants("Placemark")
        let coordinates = XDocumentHelper.GetValue(item, "Point/coordinates", "")
        select new GeoData()
        {
          SearchAddress = address,
          Status = status,
          Id = XDocumentHelper.GetValue(item, 
            "", "id"),
          Address = XDocumentHelper.GetValue(item, 
            "address", ""),
          Accuracy = (GeoAddressAccuracy)Convert.ToInt16(XDocumentHelper.GetValue(item, 
            "AddressDetails", "Accuracy")),
          Coordinate = new Coordinate()
          {
            Latitude = Convert.ToSingle(coordinates.Split(',')[1], provider),
            Longitude = Convert.ToSingle(coordinates.Split(',')[0], provider)
          },
          CountryNameCode = XDocumentHelper.GetValue(item, 
            "AddressDetails/Country/CountryNameCode", ""),
          AdministrativeAreaName = XDocumentHelper.GetValue(item, 
            "AddressDetails/Country/AdministrativeArea/AdministrativeAreaName", ""),
          SubAdministrativeAreaName = XDocumentHelper.GetValue(item, 
            "AddressDetails/Country/AdministrativeArea/SubAdministrativeArea/SubAdministrativeAreaName", ""),
          LocalityName = XDocumentHelper.GetValue(item,
            new string[] 
              {
                "AddressDetails/Country/AdministrativeArea/Locality/LocalityName",
                "AddressDetails/Country/AdministrativeArea/SubAdministrativeArea/Locality/LocalityName"
              }, ""),
          DependentLocalityName = XDocumentHelper.GetValue(item, 
            new string[] 
              {
                "AddressDetails/Country/AdministrativeArea/Locality/DependentLocality/DependentLocalityName",
                "AddressDetails/Country/AdministrativeArea/SubAdministrativeArea/Locality/DependentLocality/DependentLocalityName"
              }, ""),
          ThoroughfareName = XDocumentHelper.GetValue(item, 
            new string[] 
              {
                "AddressDetails/Country/AdministrativeArea/Locality/Thoroughfare/ThoroughfareName",
                "AddressDetails/Country/AdministrativeArea/Locality/DependentLocality/Thoroughfare/ThoroughfareName",
                "AddressDetails/Country/AdministrativeArea/SubAdministrativeArea/Locality/Thoroughfare/ThoroughfareName",
                "AddressDetails/Country/AdministrativeArea/SubAdministrativeArea/Locality/DependentLocality/Thoroughfare/ThoroughfareName"
              }, ""),
          PremiseName = XDocumentHelper.GetValue(item, 
            new string[] 
              {
                "AddressDetails/Country/AdministrativeArea/Locality/Thoroughfare/Premise/PremiseName",
                "AddressDetails/Country/AdministrativeArea/Locality/DependentLocality/Thoroughfare/Premise/PremiseName",
                "AddressDetails/Country/AdministrativeArea/SubAdministrativeArea/Locality/Thoroughfare/Thoroughfare/Premise/PremiseName",
                "AddressDetails/Country/AdministrativeArea/SubAdministrativeArea/Locality/DependentLocality/Thoroughfare/Premise/PremiseName"
              }, ""),
          PostalCodeNumber = XDocumentHelper.GetValue(item,
            new string[] 
              {
                "AddressDetails/Country/AdministrativeArea/Locality/PostalCode/PostalCodeNumber",
                "AddressDetails/Country/AdministrativeArea/Locality/DependentLocality/PostalCode/PostalCodeNumber",
                "AddressDetails/Country/AdministrativeArea/SubAdministrativeArea/Locality/PostalCode/PostalCodeNumber",
                "AddressDetails/Country/AdministrativeArea/SubAdministrativeArea/Locality/DependentLocality/PostalCode/PostalCodeNumber"
              }, ""),
        };

      return geoList.OfType<IGeoData>();
    }

    /// <summary>
    /// Asynchronous call of Google Maps geocoding service via HTTP. Pass address to Google Maps server. 
    /// GeoListCompleted event will be triggered when a collection of IGeoData objects is returned from server.
    /// </summary>
    /// <param name="address">String with as many address parameters as you want (street, housenumber, postcode, city, country, ...)</param>
    /// <example><code>
    /// GoogleGeoCoder geoCoder = new GoogleGeoCoder("A Google Maps API key");
    /// geoCoder.GeoListCompleted += new GeoListCompletedEventHandler(geoCoder_GeoListCompleted);
    /// geoCoder.GetGeoListAsync("Street City Country");
    ///
    /// void geoCoder_GeoListCompleted(object sender, IEnumerable&gt;IGeoData&lt; geoList)
    /// {
    ///   if ((geoList.Count() > 0) &amp;&amp; (geoList.First().Status == GeoStatusCode.Success))
    ///   {
    ///     foreach (var geoData in geoList)
    ///     {
    ///       Console.WriteLine("Address = {0}", geoData.Address);
    ///       Console.WriteLine("Country = {0}", geoData.CountryNameCode);
    ///       Console.WriteLine("Coordinates = {0}:{1}", geoData.Coordinate.Longitude, geoData.Coordinate.Latitude);
    ///     }
    ///   }
    /// }        
    /// </code></example>
    public void GetGeoListAsync(string address)
    {
      // Fill in arguments of the URI
      Uri baseAddressUri = new Uri("http://maps.google.com/");
      UriTemplate uriTemplate = new UriTemplate("maps/geo?q={address}&output={output}&key={key}");
      Uri formattedUri = uriTemplate.BindByPosition(baseAddressUri, address, "kml", googleMapsKey);

      // Call REST service and download XML
      WebClient webClient = new WebClient();
      webClient.DownloadDataCompleted += webClient_DownloadDataCompleted;
      webClient.DownloadDataAsync(formattedUri);
    }

    /// <summary>
    /// DownloadDataCompleted event from asynchronous call to Google Maps HTTP server
    /// </summary>
    /// <param name="sender">Sender</param>
    /// <param name="e">Event arguments</param>
    private void webClient_DownloadDataCompleted(object sender, DownloadDataCompletedEventArgs e)
    {
      if (e.Error == null)
      {
        // Convert array of bytes to string
        Byte[] bytes = e.Result;
        System.Text.ASCIIEncoding encoder = new System.Text.ASCIIEncoding();
        string geoDataXml = encoder.GetString(bytes);

        // Process Xml data into collection of IGeoData objects
        var geoList = ProcessGeoDataXml(geoDataXml, "");

        // Trigger GeoListCompleted event
        if (GeoListCompleted != null)
          GeoListCompleted(this, geoList);
      }
    }
  }
}


